###########################
# Neural Laplace: Learning diverse classes of differential equations in the Laplace domain
# Author: Samuel Holt
###########################
import torch

# Internal functions


def spherical_riemann_to_complex(theta, phi):
    r"""Spherical Riemann stereographic projection coordinates to complex number coordinates. I.e. inverse Spherical Riemann stereographic projection map.

    The inverse transform, :math:`v: \mathcal{D} \rightarrow \mathbb{C}`, is given as

    .. math::
        \begin{aligned}
            s = v(\theta, \phi) = \tan \left( \frac{\phi}{2} + \frac{\pi}{4} \right) e^{i \theta}
        \end{aligned}

    Args:
        theta (Tensor): Spherical Riemann stereographic projection coordinates, shape :math:`\theta` component of shape :math:`(\text{dimension})`.
        phi (Tensor): Spherical Riemann stereographic projection coordinates, shape :math:`\phi` component of shape :math:`(\text{dimension})`.

    Returns:
        Tuple Tensor of real and imaginary components of the complex numbers coordinate, of :math:`(\Re(s), \Im(s))`. Where :math:`s \in \mathbb{C}^d`, with :math:`d` is :math:`\text{dimension}`.

    """
    # dim(phi) == dim(theta), takes in spherical co-ordinates returns comlex real & imaginary parts
    assert phi.shape == theta.shape
    r = torch.tan(phi / 2 + torch.pi / 4)
    s_real = r * torch.cos(theta)
    s_imag = r * torch.sin(theta)
    return s_real, s_imag


def complex_to_spherical_riemann(s_real, s_imag):
    r"""Complex coordinates to to Spherical Riemann stereographic projection coordinates.
    I.e. we can translate any complex number :math:`s\in \mathbb{C}` into a coordinate on the Riemann Sphere :math:`(\theta, \phi) \in \mathcal{D} = (-{\pi}, {\pi}) \times (-\frac{\pi}{2}, \frac{\pi}{2})`, i.e.

    .. math::
        \begin{aligned}
            u(s) = \left( \arctan \left( \frac{\Im(s)}{\Re(s)} \right),\arcsin \left( \frac{|s|^2-1}{|s|^2+1} \right) \right)
        \end{aligned}

    For more details see `[1] <https://arxiv.org/abs/2206.04843>`__.

    Args:
        s_real (Tensor): Real component of the complex tensor, of shape :math:`(\text{dimension})`.
        s_imag (Tensor): Imaginary component of the complex tensor, of shape :math:`(\text{dimension})`.

    Returns:
        Tuple Tensor of :math:`(\theta, \phi)` of complex number in spherical Riemann stereographic projection coordinates. Where the shape  of :math:`\theta, \phi` is of shape :math:`(\text{dimension})`.

    """

    # din(s_real) == dim(s_imag), takes in real & complex parts returns spherical co-ordinates
    assert s_real.shape == s_imag.shape
    s_abs_2 = s_imag**2 + s_real**2
    # Handle points at infinity
    phi_r_int = torch.where(
        torch.isinf(s_abs_2), torch.ones_like(s_abs_2), ((s_abs_2 - 1) / (s_abs_2 + 1))
    )
    phi_r = torch.asin(phi_r_int)
    theta_r = torch.atan2(s_imag, s_real)
    return theta_r, phi_r


# Main functions


def spherical_to_complex(theta, phi):
    r"""Spherical Riemann stereographic projection coordinates to complex number coordinates. I.e. inverse Spherical Riemann stereographic projection map.

    The inverse transform, :math:`v: \mathcal{D} \rightarrow \mathbb{C}`, is given as

    .. math::
        \begin{aligned}
            s = v(\theta, \phi) = \tan \left( \frac{\phi}{2} + \frac{\pi}{4} \right) e^{i \theta}
        \end{aligned}

    This uses :meth:`torchlaplace.transformations.spherical_riemann_to_complex`, however provides maintains the shape of the input tensor and output tensor.

    Args:
        theta (Tensor): Spherical Riemann stereographic projection coordinates, shape :math:`\theta` component of shape :math:`(\text{Shape})`.
        phi (Tensor): Spherical Riemann stereographic projection coordinates, shape :math:`\phi` component of shape :math:`(\text{Shape})`.

    Returns:
        Complex Tensor of the complex numbers coordinate, of shape :math:`(\text{Shape})`.

    """
    # Riemann spherical to complex: R^2 -> C^1
    # dim(phi) == dim(theta), takes in spherical co-ordinates returns comlex real & imaginary parts
    s_real, s_imag = spherical_riemann_to_complex(
        torch.flatten(theta), torch.flatten(phi)
    )
    s = torch.view_as_complex(torch.stack((s_real, s_imag), 1))
    return torch.reshape(s, theta.shape)


def complex_to_spherical(s):
    r"""Complex coordinates to to Spherical Riemann stereographic projection coordinates.
    I.e. we can translate any complex number :math:`s\in \mathbb{C}` into a coordinate on the Riemann Sphere :math:`(\theta, \phi) \in \mathcal{D} = (-{\pi}, {\pi}) \times (-\frac{\pi}{2}, \frac{\pi}{2})`, i.e.

    .. math::
        \begin{aligned}
            u(s) = \left( \arctan \left( \frac{\Im(s)}{\Re(s)} \right),\arcsin \left( \frac{|s|^2-1}{|s|^2+1} \right) \right)
        \end{aligned}

    For more details see `[1] <https://arxiv.org/abs/2206.04843>`__.

    This uses :meth:`torchlaplace.transformations.complex_to_spherical_riemann`, however provides maintains the shape of the input tensor and output tensor.

    Args:
        s (Tensor): Complex tensor, of shape :math:`(\text{Shape})`.

    Returns:
        Tuple Tensor of :math:`(\theta, \phi)` of complex number in spherical Riemann stereographic projection coordinates. Where the shape of :math:`\theta, \phi` each is of shape :math:`(\text{Shape})`.

    """
    # Complex to Riemann spherical: C^1 -> R^2
    # dim(phi) == dim(theta), takes in spherical co-ordinates returns comlex real & imaginary parts
    theta, phi = complex_to_spherical_riemann(
        torch.flatten(s.real), torch.flatten(s.imag)
    )
    theta = torch.reshape(theta, s.real.shape)
    phi = torch.reshape(phi, s.imag.shape)
    return theta, phi


if __name__ == "__main__":
    angle_samples = 32
    EPS = 1e-6  # or 7, 8 or higher and numerical error issues
    phi = torch.linspace(
        -torch.pi / 2 + EPS, torch.pi / 2 - EPS, angle_samples
    ).double()
    print("phi ", phi)
    theta = torch.linspace(0 + EPS, torch.pi - EPS, angle_samples).double()
    print("theta ", theta)

    s_real, s_imag = spherical_riemann_to_complex(theta, phi)

    print("s_real ", s_real)
    print("s_imag ", s_imag)

    theta_r, phi_r = complex_to_spherical_riemann(s_real, s_imag)

    print("theta_r ", theta_r)
    print("phi_r ", phi_r)

    print("Seperate tests\n\n")

    # Inf point
    print("Point at infty")
    s_real = torch.Tensor([torch.inf]).double()
    s_imag = torch.Tensor([torch.inf]).double()
    theta_r, phi_r = complex_to_spherical_riemann(s_real, s_imag)

    print("theta_r ", theta_r)
    print("phi_r ", phi_r)

    # # Zero point
    print("Point at zero")
    s_real = torch.Tensor([0]).double()
    s_imag = torch.Tensor([0]).double()
    theta_r, phi_r = complex_to_spherical_riemann(s_real, s_imag)

    print("theta_r ", theta_r)
    print("phi_r ", phi_r)



# Ref: Neural Flows
from pathlib import Path

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Independent, kl_divergence
from torch.distributions.normal import Normal

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DATA_DIR = Path("./data")
LOCAL = False


def basic_collate_fn(
    batch,
    time_steps,
    extrap=False,
    data="",
    device=DEVICE,
    data_type="train",
    observe_step=1,
    predict_step=1,
):
    batch = torch.stack(batch)
    data_dict = {
        "data": batch,
        "time_steps": time_steps.unsqueeze(0),
    }
    data_dict = split_and_subsample_batch(
        data_dict,
        extrap=extrap,
        data=data,
        data_type=data_type,
        observe_step=observe_step,
        predict_step=predict_step,
    )
    return data_dict


def variable_time_collate_fn(
    batch,
    extrap=0,
    data="",
    device=DEVICE,
    data_type="train",
    data_min=None,
    data_max=None,
):
    """
    Expects a batch of time series data in the form of (record_id, tt, vals, mask, labels) where
            - record_id is a patient id
            - tt is a 1-dimensional tensor containing T time values of observations.
            - vals is a (T, D) tensor containing observed values for D variables.
            - mask is a (T, D) tensor containing 1 where values were observed and 0 otherwise.
            - labels is a list of labels for the current patient, if labels are available. Otherwise None.
    Returns:
            combined_tt: The union of all time observations.
            combined_vals: (M, T, D) tensor containing the observed values.
            combined_mask: (M, T, D) tensor containing 1 where values were observed and 0 otherwise.
    """
    D = batch[0][2].shape[1]

    lens = [len(x[1]) for x in batch]
    max_len = max(lens)

    combined_tt = torch.stack(
        [torch.cat([x[1], torch.zeros(max_len - l)], 0) for x, l in zip(batch, lens)], 0
    ).double()
    combined_vals = torch.stack(
        [
            torch.cat([x[2], torch.zeros(max_len - l, D)], 0)
            for x, l in zip(batch, lens)
        ],
        0,
    ).double()
    combined_mask = torch.stack(
        [
            torch.cat([x[3], torch.zeros(max_len - l, D)], 0)
            for x, l in zip(batch, lens)
        ],
        0,
    ).double()
    combined_labels = (
        torch.stack(
            [torch.tensor(float("nan")) if x[4] is None else x[4] for x in batch], 0
        )
        .to(device)
        .double()
    )
    combined_labels = combined_labels.unsqueeze(-1)

    # Normalize data an time
    combined_vals, _, _ = normalize_masked_data(
        combined_vals, combined_mask, data_min, data_max
    )
    combined_tt = combined_tt / 48.0

    data_dict = {
        "data": combined_vals,  # (batch, sequence length, dim=41)
        "time_steps": combined_tt,  # (batch, sequence length)
        "mask": combined_mask,  # (batch, sequence length, dim=41)
        "labels": combined_labels,  # (batch, 1)
    }

    data_dict = split_and_subsample_batch(data_dict, extrap=extrap, data_type=data_type)
    return data_dict


def split_and_subsample_batch(
    data_dict, extrap=False, data="", data_type="train", observe_step=1, predict_step=1
):
    if data_type == "train":
        # Training set
        if extrap:
            processed_dict = split_data_extrap(
                data_dict,
                dataset=data,
                observe_step=observe_step,
                predict_step=predict_step,
            )
        else:
            processed_dict = split_data_interp(data_dict)
    else:
        # Test set
        if extrap:
            processed_dict = split_data_extrap(
                data_dict,
                dataset=data,
                observe_step=observe_step,
                predict_step=predict_step,
            )
        else:
            processed_dict = split_data_interp(data_dict)

    # add mask
    processed_dict = add_mask(processed_dict)
    return processed_dict


def split_last_dim(data):
    last_dim = data.size()[-1]
    last_dim = last_dim // 2

    if len(data.size()) == 3:
        res = data[:, :, :last_dim], data[:, :, last_dim:]

    if len(data.size()) == 2:
        res = data[:, :last_dim], data[:, last_dim:]
    return res


def init_network_weights(net, std=0.1):
    for m in net.modules():
        if isinstance(m, nn.Linear):
            nn.init.normal_(m.weight, mean=0, std=std)
            nn.init.constant_(m.bias, val=0)


def get_device(tensor):
    device = torch.device("cpu")
    if tensor.is_cuda:
        device = tensor.get_device()
    return device


def sample_standard_gaussian(mu, sigma):
    device = get_device(mu)

    d = torch.distributions.normal.Normal(
        torch.Tensor([0.0]).to(device), torch.Tensor([1.0]).to(device)
    )
    r = d.sample(mu.size()).squeeze(-1)
    return r * sigma.float() + mu.float()


def split_train_val_test(data, train_frac=0.6, val_frac=0.2):
    n_samples = len(data)
    data_train = data[: int(n_samples * train_frac)]
    data_val = data[
        int(n_samples * train_frac) : int(n_samples * (train_frac + val_frac))
    ]
    data_test = data[int(n_samples * (train_frac + val_frac)) :]
    return data_train, data_val, data_test


def linspace_vector(start, end, n_points):
    # start is either one value or a vector
    size = np.prod(start.size())

    assert start.size() == end.size()
    if size == 1:
        # start and end are 1d-tensors
        res = torch.linspace(start, end, n_points)
    else:
        # start and end are vectors
        res = torch.Tensor()
        for i in range(0, start.size(0)):
            res = torch.cat((res, torch.linspace(start[i], end[i], n_points)), 0)
        res = torch.t(res.reshape(start.size(0), n_points))
    return res


def reverse(tensor):
    idx = [i for i in range(tensor.size(0) - 1, -1, -1)]
    return tensor[idx]


def create_net(n_inputs, n_outputs, n_layers=1, n_units=100, nonlinear=nn.Tanh):
    layers = [nn.Linear(n_inputs, n_units)]
    for i in range(n_layers):
        layers.append(nonlinear())
        layers.append(nn.Linear(n_units, n_units))

    layers.append(nonlinear())
    layers.append(nn.Linear(n_units, n_outputs))
    return nn.Sequential(*layers)


def get_dict_template():
    return {
        "observed_data": None,
        "observed_tp": None,
        "data_to_predict": None,
        "tp_to_predict": None,
        "observed_mask": None,
        "mask_predicted_data": None,
        "labels": None,
    }


def normalize_data(data):
    reshaped = data.reshape(-1, data.size(-1))

    att_min = torch.min(reshaped, 0)[0]
    att_max = torch.max(reshaped, 0)[0]

    # we don't want to divide by zero
    att_max[att_max == 0.0] = 1.0

    if (att_max != 0.0).all():
        data_norm = (data - att_min) / att_max
    else:
        raise Exception("Zero!")

    if torch.isnan(data_norm).any():
        raise Exception("nans!")

    return data_norm, att_min, att_max


def normalize_masked_data(data, mask, att_min, att_max):
    # we don't want to divide by zero
    att_max[att_max == 0.0] = 1.0

    if (att_max != 0.0).all():
        data_norm = (data - att_min) / att_max
    else:
        raise Exception("Zero!")

    if torch.isnan(data_norm).any():
        raise Exception("nans!")

    # set masked out elements back to zero
    data_norm[mask == 0] = 0

    return data_norm, att_min, att_max


def shift_outputs(outputs, first_datapoint=None):
    outputs = outputs[:, :, :-1, :]

    if first_datapoint is not None:
        n_traj, n_dims = first_datapoint.size()
        first_datapoint = first_datapoint.reshape(1, n_traj, 1, n_dims)
        outputs = torch.cat((first_datapoint, outputs), 2)
    return outputs


def split_data_extrap(data_dict, dataset="", observe_step=1, predict_step=1):
    n_observed_tp = data_dict["data"].size(1) // 2
    if dataset == "hopper":
        n_observed_tp = data_dict["data"].size(1) // 3

    split_dict = {
        "observed_data": data_dict["data"][:, :n_observed_tp, :][
            :, ::observe_step, :
        ].clone(),
        "observed_tp": data_dict["time_steps"][:, :n_observed_tp][
            :, ::observe_step
        ].clone(),
        "data_to_predict": data_dict["data"][:, n_observed_tp:, :][
            :, ::predict_step, :
        ].clone(),
        "tp_to_predict": data_dict["time_steps"][:, n_observed_tp:][
            :, ::predict_step
        ].clone(),
    }

    split_dict["observed_mask"] = None
    split_dict["mask_predicted_data"] = None
    split_dict["labels"] = None

    if ("mask" in data_dict) and (data_dict["mask"] is not None):
        split_dict["observed_mask"] = data_dict["mask"][:, :n_observed_tp].clone()
        split_dict["mask_predicted_data"] = data_dict["mask"][:, n_observed_tp:].clone()

    if ("labels" in data_dict) and (data_dict["labels"] is not None):
        split_dict["labels"] = data_dict["labels"].clone()

    split_dict["mode"] = "extrap"
    return split_dict


def split_data_interp(data_dict):
    split_dict = {
        "observed_data": data_dict["data"].clone(),
        "observed_tp": data_dict["time_steps"].clone(),
        "data_to_predict": data_dict["data"].clone(),
        "tp_to_predict": data_dict["time_steps"].clone(),
    }

    split_dict["observed_mask"] = None
    split_dict["mask_predicted_data"] = None
    split_dict["labels"] = None

    if "mask" in data_dict and data_dict["mask"] is not None:
        split_dict["observed_mask"] = data_dict["mask"].clone()
        split_dict["mask_predicted_data"] = data_dict["mask"].clone()

    if ("labels" in data_dict) and (data_dict["labels"] is not None):
        split_dict["labels"] = data_dict["labels"].clone()

    split_dict["mode"] = "interp"
    return split_dict


def add_mask(data_dict):
    data = data_dict["observed_data"]
    mask = data_dict["observed_mask"]

    if mask is None:
        mask = torch.ones_like(data).to(get_device(data))

    data_dict["observed_mask"] = mask
    return data_dict


def compute_loss_all_batches(
    model,
    dl,
    device,
    classify=0,
    data="",
    n_traj_samples=3,
    kl_coef=1.0,
    max_samples_for_eval=None,
):
    total = {}
    total["loss"] = 0
    total["likelihood"] = 0
    total["mse"] = 0
    total["acc"] = 0
    total["kl_first_p"] = 0
    total["std_first_p"] = 0
    total["pois_likelihood"] = 0
    total["ce_loss"] = 0

    n_test_batches = 0

    classif_predictions = torch.Tensor([]).to(device)
    all_test_labels = torch.Tensor([]).to(device)

    for batch_dict in dl:
        results = model.compute_all_losses(
            batch_dict, n_traj_samples=n_traj_samples, kl_coef=kl_coef
        )

        if classify and data != "hopper":
            n_labels = model.n_labels  # batch_dict['labels'].size(-1)
            n_traj_samples = results["label_predictions"].size(0)

            classif_predictions = torch.cat(
                (
                    classif_predictions,
                    results["label_predictions"].reshape(n_traj_samples, -1, n_labels),
                ),
                1,
            )
            all_test_labels = torch.cat(
                (all_test_labels, batch_dict["labels"].reshape(-1, n_labels)), 0
            )

        for key in total.keys():
            if key in results:
                var = results[key]
                if isinstance(var, torch.Tensor):
                    var = var.detach()
                total[key] += var

        n_test_batches += 1

    if n_test_batches > 0:
        for key, _ in total.items():
            total[key] = total[key] / n_test_batches

    if classify:
        if data == "physionet":
            all_test_labels = all_test_labels.repeat(n_traj_samples, 1, 1)

            idx_not_nan = ~torch.isnan(all_test_labels)
            classif_predictions = classif_predictions[idx_not_nan]
            all_test_labels = all_test_labels[idx_not_nan]

            total["acc"] = 0.0  # AUC score
            if torch.sum(all_test_labels) != 0.0:
                print(f"Number of labeled examples: {len(all_test_labels.reshape(-1))}")
                print(
                    "Number of examples with mortality 1: {}".format(
                        torch.sum(all_test_labels == 1.0)
                    )
                )

                # Cannot compute AUC with only 1 class
                import sklearn as sk

                total["acc"] = sk.metrics.roc_auc_score(
                    all_test_labels.cpu().numpy().reshape(-1),
                    classif_predictions.cpu().numpy().reshape(-1),
                )
            else:
                print(
                    "Warning: Couldn't compute AUC -- all examples are from the same class"
                )

        if data == "activity":
            all_test_labels = all_test_labels.repeat(n_traj_samples, 1, 1)

            labeled_tp = torch.sum(all_test_labels, -1, keepdim=True) > 0.0
            labeled_tp = labeled_tp.repeat_interleave(all_test_labels.shape[-1], -1)

            all_test_labels = all_test_labels[labeled_tp].view(
                -1, all_test_labels.shape[-1]
            )
            classif_predictions = classif_predictions[labeled_tp].view(
                -1, all_test_labels.shape[-1]
            )

            # classif_predictions and all_test_labels are in on-hot-encoding -- convert to class ids
            _, pred_class_id = torch.max(classif_predictions, -1)
            _, class_labels = torch.max(all_test_labels, -1)

            total["acc"] = torch.sum(class_labels == pred_class_id).item() / sum(
                class_labels.shape
            )

    return total


def check_mask(data, mask):
    # check that 'mask' argument indeed contains a mask for data
    n_zeros = torch.sum(mask == 0.0).cpu().numpy()
    n_ones = torch.sum(mask == 1.0).cpu().numpy()

    # mask should contain only zeros and ones
    assert (n_zeros + n_ones) == np.prod(list(mask.size()))

    # all masked out elements should be zeros
    assert torch.sum(data[mask == 0.0] != 0.0) == 0


def create_classifier(z0_dim, n_labels):
    return nn.Sequential(
        nn.Linear(z0_dim, 300),
        nn.ReLU(),
        nn.Linear(300, 300),
        nn.ReLU(),
        nn.Linear(300, n_labels),
    )


class VAE_Baseline(nn.Module):
    def __init__(
        self,
        input_dim,
        latent_dim,
        z0_prior,
        device,
        obsrv_std=0.01,
        use_binary_classif=False,
        classif_per_tp=False,
        use_poisson_proc=False,
        linear_classifier=False,
        n_labels=1,
        train_classif_w_reconstr=False,
    ):

        super(VAE_Baseline, self).__init__()

        self.input_dim = input_dim
        self.latent_dim = latent_dim
        self.device = device
        self.n_labels = n_labels

        self.obsrv_std = torch.Tensor([obsrv_std]).to(device)

        self.z0_prior = z0_prior
        self.use_binary_classif = use_binary_classif
        self.classif_per_tp = classif_per_tp
        self.use_poisson_proc = use_poisson_proc
        self.linear_classifier = linear_classifier
        self.train_classif_w_reconstr = train_classif_w_reconstr

        z0_dim = latent_dim
        if use_poisson_proc:
            z0_dim += latent_dim

        if use_binary_classif:
            if linear_classifier:
                self.classifier = nn.Sequential(nn.Linear(z0_dim, n_labels))
            else:
                self.classifier = create_classifier(z0_dim, n_labels)
            init_network_weights(self.classifier)

    def get_gaussian_likelihood(self, truth, pred_y, mask=None):
        # pred_y shape [n_traj_samples, n_traj, n_tp, n_dim]
        # truth shape  [n_traj, n_tp, n_dim]
        n_traj, n_tp, n_dim = truth.size()

        # Compute likelihood of the data under the predictions
        truth_repeated = truth.repeat(pred_y.size(0), 1, 1, 1)

        if mask is not None:
            mask = mask.repeat(pred_y.size(0), 1, 1, 1)
        log_density_data = masked_gaussian_log_density(
            pred_y, truth_repeated, obsrv_std=self.obsrv_std, mask=mask
        )
        log_density_data = log_density_data.permute(1, 0)
        log_density = torch.mean(log_density_data, 1)

        # shape: [n_traj_samples]
        return log_density

    def get_mse(self, truth, pred_y, mask=None):
        # pred_y shape [n_traj_samples, n_traj, n_tp, n_dim]
        # truth shape  [n_traj, n_tp, n_dim]
        n_traj, n_tp, n_dim = truth.size()

        # Compute likelihood of the data under the predictions
        truth_repeated = truth.repeat(pred_y.size(0), 1, 1, 1)

        if mask is not None:
            mask = mask.repeat(pred_y.size(0), 1, 1, 1)

        # Compute likelihood of the data under the predictions
        log_density_data = compute_mse(pred_y, truth_repeated, mask=mask)
        # shape: [1]
        return torch.mean(log_density_data)

    def compute_all_losses(self, batch_dict, n_traj_samples=1, kl_coef=1.0):
        # Condition on subsampled points
        # Make predictions for all the points
        pred_y, info = self.get_reconstruction(
            batch_dict["tp_to_predict"],
            batch_dict["observed_data"],
            batch_dict["observed_tp"],
            mask=batch_dict["observed_mask"],
            n_traj_samples=n_traj_samples,
            mode=batch_dict["mode"],
        )

        # print('get_reconstruction done -- computing likelihood')
        fp_mu, fp_std, fp_enc = info["first_point"]
        fp_distr = Normal(fp_mu, fp_std)

        assert torch.sum(fp_std < 0) == 0.0

        kldiv_z0 = kl_divergence(fp_distr, self.z0_prior)

        if torch.isnan(kldiv_z0).any():
            print(fp_mu)
            print(fp_std)
            raise Exception("kldiv_z0 is Nan!")

        # Mean over number of latent dimensions
        # kldiv_z0 shape: [n_traj_samples, n_traj, n_latent_dims] if prior is a mixture of gaussians (KL is estimated)
        # kldiv_z0 shape: [1, n_traj, n_latent_dims] if prior is a standard gaussian (KL is computed exactly)
        # shape after: [n_traj_samples]
        kldiv_z0 = torch.mean(kldiv_z0, (1, 2))

        # Compute likelihood of all the points
        rec_likelihood = self.get_gaussian_likelihood(
            batch_dict["data_to_predict"],
            pred_y,
            mask=batch_dict["mask_predicted_data"],
        )

        mse = self.get_mse(
            batch_dict["data_to_predict"],
            pred_y,
            mask=batch_dict["mask_predicted_data"],
        )

        pois_log_likelihood = torch.Tensor([0.0]).to(batch_dict["data_to_predict"])
        if self.use_poisson_proc:
            pois_log_likelihood = compute_poisson_proc_likelihood(
                batch_dict["data_to_predict"],
                pred_y,
                info,
                mask=batch_dict["mask_predicted_data"],
            )
            # Take mean over n_traj
            pois_log_likelihood = torch.mean(pois_log_likelihood, 1)

        ################################
        # Compute CE loss for binary classification on Physionet
        ce_loss = torch.Tensor([0.0]).to(batch_dict["data_to_predict"])
        if (batch_dict["labels"] is not None) and self.use_binary_classif:

            if (batch_dict["labels"].size(-1) == 1) or (
                len(batch_dict["labels"].size()) == 1
            ):
                ce_loss = compute_binary_CE_loss(
                    info["label_predictions"], batch_dict["labels"]
                )
            else:
                ce_loss = compute_multiclass_CE_loss(
                    info["label_predictions"],
                    batch_dict["labels"],
                    mask=batch_dict["mask_predicted_data"],
                )

        # IWAE loss
        loss = -torch.logsumexp(rec_likelihood - kl_coef * kldiv_z0, 0)
        if torch.isnan(loss):
            loss = -torch.mean(rec_likelihood - kl_coef * kldiv_z0, 0)

        if self.use_poisson_proc:
            loss = loss - 0.1 * pois_log_likelihood

        if self.use_binary_classif:
            if self.train_classif_w_reconstr:
                loss = loss + ce_loss * 100
            else:
                loss = ce_loss

        results = {}
        results["loss"] = torch.mean(loss)
        results["likelihood"] = torch.mean(rec_likelihood).detach()
        results["mse"] = torch.mean(mse).detach()
        results["pois_likelihood"] = torch.mean(pois_log_likelihood).detach()
        results["ce_loss"] = torch.mean(ce_loss).detach()
        results["kl_first_p"] = torch.mean(kldiv_z0).detach()
        results["std_first_p"] = torch.mean(fp_std).detach()

        if batch_dict["labels"] is not None and self.use_binary_classif:
            results["label_predictions"] = info["label_predictions"].detach()

        return results


def get_mask(x):
    x = x.unsqueeze(0)
    n_data_dims = x.size(-1) // 2
    mask = x[:, :, n_data_dims:]
    check_mask(x[:, :, :n_data_dims], mask)
    mask = (torch.sum(mask, -1, keepdim=True) > 0).float()
    assert not torch.isnan(mask).any()
    return mask.squeeze(0)


class Encoder_z0_ODE_RNN(nn.Module):
    def __init__(
        self,
        latent_dim,
        input_dim,
        z0_diffeq_solver=None,
        z0_dim=None,
        n_gru_units=100,
        device=torch.device("cpu"),
    ):
        super().__init__()

        if z0_dim is None:
            self.z0_dim = latent_dim
        else:
            self.z0_dim = z0_dim

        self.lstm = nn.LSTMCell(input_dim, latent_dim)

        self.z0_diffeq_solver = z0_diffeq_solver
        self.latent_dim = latent_dim
        self.input_dim = input_dim
        self.device = device
        self.extra_info = None

        self.transform_z0 = nn.Sequential(
            nn.Linear(latent_dim, 100),
            nn.Tanh(),
            nn.Linear(100, self.z0_dim * 2),
        )
        init_network_weights(self.transform_z0)

    def forward(self, data, time_steps, run_backwards=True, save_info=False):
        assert not torch.isnan(data).any()
        assert not torch.isnan(time_steps).any()

        n_traj, n_tp, n_dims = data.size()
        latent = self.run_odernn(data, time_steps, run_backwards)

        latent = latent.reshape(1, n_traj, self.latent_dim)

        mean_z0, std_z0 = self.transform_z0(latent).chunk(2, dim=-1)
        std_z0 = F.softplus(std_z0)

        return mean_z0, std_z0

    def run_odernn(self, data, time_steps, run_backwards=True):
        batch_size, n_tp, n_dims = data.size()
        prev_t, t_i = time_steps[:, -1] + 0.01, time_steps[:, -1]

        time_points_iter = range(0, time_steps.shape[1])
        if run_backwards:
            time_points_iter = reversed(time_points_iter)

        h = torch.zeros(batch_size, self.latent_dim).to(data)
        c = torch.zeros(batch_size, self.latent_dim).to(data)

        for i in time_points_iter:
            t = (t_i - prev_t).unsqueeze(1)
            h = self.z0_diffeq_solver(h.unsqueeze(1), t).squeeze(1)

            xi = data[:, i, :]
            h_, c_ = self.lstm(xi, (h, c))
            mask = get_mask(xi)

            h = mask * h_ + (1 - mask) * h
            c = mask * c_ + (1 - mask) * c

            prev_t, t_i = time_steps[:, i], time_steps[:, i - 1]

        return h


class Decoder(nn.Module):
    def __init__(self, latent_dim, input_dim):
        super().__init__()
        decoder = nn.Sequential(
            nn.Linear(latent_dim, input_dim),
        )
        init_network_weights(decoder)
        self.decoder = decoder

    def forward(self, data):
        return self.decoder(data)


def gaussian_log_likelihood(mu_2d, data_2d, obsrv_std, indices=None):
    n_data_points = mu_2d.size()[-1]

    if n_data_points > 0:
        gaussian = Independent(
            Normal(loc=mu_2d, scale=obsrv_std.repeat(n_data_points)), 1
        )
        log_prob = gaussian.log_prob(data_2d)
        log_prob = log_prob / n_data_points
    else:
        log_prob = torch.zeros([1]).to(data_2d).squeeze()
    return log_prob


def poisson_log_likelihood(masked_log_lambdas, masked_data, indices, int_lambdas):
    # masked_log_lambdas and masked_data
    n_data_points = masked_data.size()[-1]

    if n_data_points > 0:
        log_prob = torch.sum(masked_log_lambdas) - int_lambdas[indices]
        # log_prob = log_prob / n_data_points
    else:
        log_prob = torch.zeros([1]).to(masked_data).squeeze()
    return log_prob


def compute_binary_CE_loss(label_predictions, mortality_label):
    # print('Computing binary classification loss: compute_CE_loss')

    mortality_label = mortality_label.reshape(-1)

    if len(label_predictions.size()) == 1:
        label_predictions = label_predictions.unsqueeze(0)

    n_traj_samples = label_predictions.size(0)
    label_predictions = label_predictions.reshape(n_traj_samples, -1)

    idx_not_nan = ~torch.isnan(mortality_label)
    if len(idx_not_nan) == 0.0:
        print("All are labels are NaNs!")
        ce_loss = torch.Tensor(0.0).to(mortality_label)

    label_predictions = label_predictions[:, idx_not_nan]
    mortality_label = mortality_label[idx_not_nan]

    if torch.sum(mortality_label == 0.0) == 0 or torch.sum(mortality_label == 1.0) == 0:
        print(
            "Warning: all examples in a batch belong to the same class -- please increase the batch size."
        )

    assert not torch.isnan(label_predictions).any()
    assert not torch.isnan(mortality_label).any()

    # For each trajectory, we get n_traj_samples samples from z0 -- compute loss on all of them
    mortality_label = mortality_label.repeat(n_traj_samples, 1)
    ce_loss = nn.BCEWithLogitsLoss()(label_predictions, mortality_label)

    # divide by number of patients in a batch
    ce_loss = ce_loss / n_traj_samples
    return ce_loss


def compute_multiclass_CE_loss(label_predictions, true_label, mask):
    # print('Computing multi-class classification loss: compute_multiclass_CE_loss')

    if len(label_predictions.size()) == 3:
        label_predictions = label_predictions.unsqueeze(0)

    n_traj_samples, n_traj, n_tp, n_dims = label_predictions.size()

    # assert(not torch.isnan(label_predictions).any())
    # assert(not torch.isnan(true_label).any())

    # For each trajectory, we get n_traj_samples samples from z0 -- compute loss on all of them
    true_label = true_label.repeat(n_traj_samples, 1, 1)

    label_predictions = label_predictions.reshape(
        n_traj_samples * n_traj * n_tp, n_dims
    )
    true_label = true_label.reshape(n_traj_samples * n_traj * n_tp, n_dims)

    # choose time points with at least one measurement
    mask = torch.sum(mask, -1) > 0

    # repeat the mask for each label to mark that the label for this time point is present
    pred_mask = mask.repeat(n_dims, 1, 1).permute(1, 2, 0)

    label_mask = mask
    pred_mask = pred_mask.repeat(n_traj_samples, 1, 1, 1)
    label_mask = label_mask.repeat(n_traj_samples, 1, 1, 1)

    pred_mask = pred_mask.reshape(n_traj_samples * n_traj * n_tp, n_dims)
    label_mask = label_mask.reshape(n_traj_samples * n_traj * n_tp, 1)

    if (label_predictions.size(-1) > 1) and (true_label.size(-1) > 1):
        assert label_predictions.size(-1) == true_label.size(-1)
        # targets are in one-hot encoding -- convert to indices
        _, true_label = true_label.max(-1)

    res = []
    for i in range(true_label.size(0)):
        pred_masked = torch.masked_select(label_predictions[i], pred_mask[i].bool())
        labels = torch.masked_select(true_label[i], label_mask[i].bool())

        pred_masked = pred_masked.reshape(-1, n_dims)

        if len(labels) == 0:
            continue

        ce_loss = nn.CrossEntropyLoss()(pred_masked, labels.long())
        res.append(ce_loss)

    ce_loss = torch.stack(res, 0).to(label_predictions)
    ce_loss = torch.mean(ce_loss)
    # # divide by number of patients in a batch
    # ce_loss = ce_loss / n_traj_samples
    return ce_loss


def compute_masked_likelihood(mu, data, mask, likelihood_func):
    # Compute the likelihood per patient and per attribute so that we don't priorize patients with more measurements
    n_traj_samples, n_traj, n_timepoints, n_dims = data.size()

    res = []
    for i in range(n_traj_samples):
        for k in range(n_traj):
            for j in range(n_dims):
                data_masked = torch.masked_select(
                    data[i, k, :, j], mask[i, k, :, j].bool()
                )

                # assert(torch.sum(data_masked == 0.) < 10)

                mu_masked = torch.masked_select(mu[i, k, :, j], mask[i, k, :, j].bool())
                log_prob = likelihood_func(mu_masked, data_masked, indices=(i, k, j))
                res.append(log_prob)
    # shape: [n_traj*n_traj_samples, 1]

    res = torch.stack(res, 0).to(data)
    res = res.reshape((n_traj_samples, n_traj, n_dims))
    # Take mean over the number of dimensions
    res = torch.mean(res, -1)  # !!!!!!!!!!! changed from sum to mean
    res = res.transpose(0, 1)
    return res


def masked_gaussian_log_density(mu, data, obsrv_std, mask=None):
    # these cases are for plotting through plot_estim_density
    if len(mu.size()) == 3:
        # add additional dimension for gp samples
        mu = mu.unsqueeze(0)

    if len(data.size()) == 2:
        # add additional dimension for gp samples and time step
        data = data.unsqueeze(0).unsqueeze(2)
    elif len(data.size()) == 3:
        # add additional dimension for gp samples
        data = data.unsqueeze(0)

    n_traj_samples, n_traj, n_timepoints, n_dims = mu.size()

    assert data.size()[-1] == n_dims

    # Shape after permutation: [n_traj, n_traj_samples, n_timepoints, n_dims]
    if mask is None:
        mu_flat = mu.reshape(n_traj_samples * n_traj, n_timepoints * n_dims)
        n_traj_samples, n_traj, n_timepoints, n_dims = data.size()
        data_flat = data.reshape(n_traj_samples * n_traj, n_timepoints * n_dims)

        res = gaussian_log_likelihood(mu_flat, data_flat, obsrv_std)
        res = res.reshape(n_traj_samples, n_traj).transpose(0, 1)
    else:
        # Compute the likelihood per patient so that we don't priorize patients with more measurements
        def func(mu, data, indices):
            return gaussian_log_likelihood(
                mu, data, obsrv_std=obsrv_std, indices=indices
            )

        res = compute_masked_likelihood(mu, data, mask, func)
    return res


def mse(mu, data, indices=None):
    n_data_points = mu.size()[-1]

    if n_data_points > 0:
        mse = nn.MSELoss()(mu, data)
    else:
        mse = torch.zeros([1]).to(data).squeeze()
    return mse


def compute_mse(mu, data, mask=None):
    # these cases are for plotting through plot_estim_density
    if len(mu.size()) == 3:
        # add additional dimension for gp samples
        mu = mu.unsqueeze(0)

    if len(data.size()) == 2:
        # add additional dimension for gp samples and time step
        data = data.unsqueeze(0).unsqueeze(2)
    elif len(data.size()) == 3:
        # add additional dimension for gp samples
        data = data.unsqueeze(0)

    n_traj_samples, n_traj, n_timepoints, n_dims = mu.size()
    assert data.size()[-1] == n_dims

    # Shape after permutation: [n_traj, n_traj_samples, n_timepoints, n_dims]
    if mask is None:
        mu_flat = mu.reshape(n_traj_samples * n_traj, n_timepoints * n_dims)
        n_traj_samples, n_traj, n_timepoints, n_dims = data.size()
        data_flat = data.reshape(n_traj_samples * n_traj, n_timepoints * n_dims)
        res = mse(mu_flat, data_flat)
    else:
        # Compute the likelihood per patient so that we don't priorize patients with more measurements
        res = compute_masked_likelihood(mu, data, mask, mse)
    return res


def compute_poisson_proc_likelihood(truth, pred_y, info, mask=None):
    # Compute Poisson likelihood
    # https://math.stackexchange.com/questions/344487/log-likelihood-of-a-realization-of-a-poisson-process
    # Sum log lambdas across all time points
    if mask is None:
        poisson_log_l = torch.sum(info["log_lambda_y"], 2) - info["int_lambda"]
        # Sum over data dims
        poisson_log_l = torch.mean(poisson_log_l, -1)
    else:
        # Compute likelihood of the data under the predictions
        truth_repeated = truth.repeat(pred_y.size(0), 1, 1, 1)
        mask_repeated = mask.repeat(pred_y.size(0), 1, 1, 1)

        # Compute the likelihood per patient and per attribute so that we don't priorize patients with more measurements
        int_lambda = info["int_lambda"]

        def f(log_lam, data, indices):
            return poisson_log_likelihood(log_lam, data, indices, int_lambda)

        poisson_log_l = compute_masked_likelihood(
            info["log_lambda_y"], truth_repeated, mask_repeated, f
        )
        poisson_log_l = poisson_log_l.permute(1, 0)
        # Take mean over n_traj
        # poisson_log_l = torch.mean(poisson_log_l, 1)

    # poisson_log_l shape: [n_traj_samples, n_traj]
    return poisson_log_l



###########################
# Neural Laplace: Learning diverse classes of differential equations in the Laplace domain
# Author: Samuel Holt
###########################
import sys
from time import time

import numpy as np
import torch
from scipy.special import factorial
from torch import nn

try:
    from ._iltcme import cme_params_factory
except ImportError:
    from _iltcme import cme_params_factory  # For when running this single file as main

EPS = sys.float_info.epsilon
TORCH_FLOAT_DATATYPE = torch.float32
TORCH_COMPLEX_DATATYPE = torch.cfloat

device = "cuda" if torch.cuda.is_available() else "cpu"


def real_vector_to_complex(vector):
    if "ComplexDoubleTensor" not in vector.type():
        if len(vector.shape) == 1:
            return torch.view_as_complex(
                torch.stack((vector, torch.zeros_like(vector)), 1)
            )
        else:
            return torch.reshape(
                torch.view_as_complex(
                    torch.stack(
                        (
                            torch.flatten(vector),
                            torch.zeros_like(torch.flatten(vector)),
                        ),
                        1,
                    )
                ),
                vector.shape,
            )
    else:
        return vector


def complex_numpy_to_complex_torch(sn):
    return torch.view_as_complex(
        torch.stack(
            (
                torch.flatten(torch.Tensor(sn.real)),
                torch.flatten(torch.Tensor(sn.imag)),
            ),
            1,
        )
    )


class InverseLaplaceTransformAlgorithmBase(nn.Module):
    r"""Base class for Inverse Laplace Transform (ILT) Algorithms. This reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.

    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`,

    .. math::
        \begin{aligned}
            \mathbf{x}(t) & = \text{inverse_laplace_transform}(\mathbf{F}(\mathbf{s}), t)
        \end{aligned}

    Where :math:`t` is the time points to reconstruct trajectories for.

    The Laplace representation function :attr:`fs`, :math:`\mathbf{F}(\mathbf{s})` takes an input complex value :math:`\mathbf{s}`.
    This :math:`\mathbf{s}` is used internally when reconstructing a specified time point with the selected inverse Laplace transform algorithm class.

    The reconstructions :math:`\mathbf{x}(t)` are of shape :math:`(\text{MiniBatchSize}, \text{SeqLen}, d_{\text{obs}})`.
    Where :math:`\text{SeqLen}` dimension corresponds to the input evaluated time points :math:`t`, and :math:`d_{\text{obs}}` is the trajectory dimension for a given time point, :math:`\mathbf{x}(t)`.

    The inverse Laplace transform (ILT) is defined as

    .. math::
        \begin{aligned}
            \hat{\mathbf{x}}(t) = \mathcal{L}^{-1}\{\mathbf{F}(\mathbf{s})\}(t)=\frac{1}{2\pi i} \int_{\sigma - i \infty}^{\sigma + i \infty} \mathbf{F}(\mathbf{s})e^{\mathbf{s}t}d\mathbf{s}
        \end{aligned}

    where the integral refers to the Bromwich contour integral in :math:`\mathbb{C}^d` with the contour :math:`\sigma>0` chosen such that all the singularities of :math:`\mathbf{F}(\mathbf{s})`
    are to the left of it `[1] <https://arxiv.org/abs/2206.04843>`__.

    Many algorithms have been developed to numerically evaluate the ILT Equation (above). On a high level, they involve two steps:

    .. math::
        \begin{aligned}
            \mathcal{Q}(t) &= \text{ILT-Query} (t) \\
            \hat{\mathbf{x}}(t) &= \text{ILT-Compute}\big(\{\mathbf{F}(\mathbf{s})| \mathbf{s} \in \mathcal{Q}(t) \}\big)
        \end{aligned}

    To evaluate :math:`\mathbf{x}(t)` on time points :math:`t \in \mathcal{T} \subset \mathbb{R}`, the algorithms first construct a set of
    `query points` :math:`\mathbf{s} \in \mathcal{Q}(\mathcal{T}) \subset \mathbb{C}`.
    They then compute :math:`\hat{\mathbf{x}}(t)` using the :math:`\mathbf{F}(\mathbf{s})` evaluated on these points.
    The number of query points scales `linearly` with the number of time points, i.e. :math:`|\mathcal{Q}(\mathcal{T})| = b |\mathcal{T}|`, where the constant :math:`b > 1`,
    denotes the number of reconstruction terms per time point and is specific to the algorithm.
    Importantly, the computation complexity of ILT only depends on the `number` of time points, but not their values (e.g. ILT for :math:`t=0` and :math:`t=100` requires the same amount of computation).
    The vast majority of ILT algorithms are differentiable with respect to :math:`\mathbf{F}(\mathbf{s})`, which allows the gradients to be back propagated through the ILT transform `[1] <https://arxiv.org/abs/2206.04843>`__.

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        super(InverseLaplaceTransformAlgorithmBase, self).__init__()
        self.ilt_reconstruction_terms = ilt_reconstruction_terms
        self.torch_float_datatype = torch_float_datatype
        self.torch_complex_datatype = torch_complex_datatype

    def forward(self, fs, ti):
        r"""Reconstructs a trajectory :math:`\mathbf{x}(t)` for a Laplace representation :math:`\mathbf{F}(\mathbf{s})`, at time points :math:`t`.

        Args:
            fs (Torch.nn.Module or Callable): The first parameter.
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`.
            :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`.
            This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        raise NotImplementedError(
            "Forward method is not implemented yet for this ILT algorithm"
        )

    def compute_s(self, ti):
        r"""Computes :math:`\mathbf{s}` to evaluate the Laplace representation :math:`\mathbf{F}(\mathbf{s})` at, from the input time points :math:`t`, using the selected ILT algorithm.

        Args:
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.

        Returns:
            Tensor of complex s points :math:`\mathbf{s}` of shape :math:`(\text{SeqLen}, \text{ReconTerms})`. :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`.

        """
        raise NotImplementedError(
            "compute_s method is not implemented yet for this ILT algorithm"
        )

    def line_integrate(self, fp, ti):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        raise NotImplementedError(
            "line_integrate method is not implemented yet for this ILT algorithm"
        )

    def line_integrate_all_multi(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm (takes batch input of `fp`).

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`.
            Best practice for most ILT algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}})`.
            :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # Input shapes
        # fp [batch, times, data_dims, approximation_terms * 2 + 1]
        # ti [times]
        # T [times]
        # Returns
        # [batch, times, data_dims]
        raise NotImplementedError(
            "line_integrate_all_multi method is not implemented yet for this ILT algorithm"
        )


class FixedTablot(InverseLaplaceTransformAlgorithmBase):
    r"""Inherits from :meth:`torchlaplace.inverse_laplace.InverseLaplaceTransformAlgorithmBase`.

    Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.
    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`.

    Deforms the Bromwich contour around the negative real axis, where :math:`\mathbf{F}(\mathbf{s})` must not overflow as :math:`\mathbf{s} \to -\infty`,
    and makes the Bromwich contour integral rapidly converge as :math:`\mathbf{s} \to -\infty` causes :math:`e^{\mathbf{s}t} \to 0` in ILT Equation.
    We implemented the Fixed Tablot method, which is simple to implement.
    However it suffers from not being able to model solutions that have large sinusoidal components and instead is optimized for modelling decaying exponential solutions.
    We note that whilst it can approximate some small sinusoidal components, for an adaptive time contour, the sinusoidal components that can be represented decrease when
     modelling longer time trajectories, and in the limit for long time horizons, allow only representations of decaying exponentials.

    .. note::
        References: Abate, J. and Valko, P. P. Multi-precision laplace transform inversion. International Journal for Numerical Methods in Engineering, 60:979–993, 2004.

        Talbot, A. The accurate numerical inversion of laplace transforms. IMA Journal of Applied Mathematics, 23(1): 97–120, 1979.

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        super().__init__(
            ilt_reconstruction_terms, torch_float_datatype, torch_complex_datatype
        )
        M = int((ilt_reconstruction_terms - 1) / 2)
        self.M = M
        self.rdt = torch.Tensor([(2.0 / 5.0) * self.M]).to(device)
        k = torch.arange(1, M, dtype=torch_float_datatype, device=torch.device(device))
        self.theta = (k * torch.pi) / M
        self.sdt = self.rdt * self.theta * (1 / torch.tan(self.theta) + 1j)
        self.torch_float_datatype = torch_float_datatype

    def forward(self, fs, ti, time_max=None):
        r"""Reconstructs a trajectory :math:`\mathbf{x}(t)` for a Laplace representation :math:`\mathbf{F}(\mathbf{s})`, at time points :math:`t`.

        Args:
            fs (Torch.nn.Module or Callable): The first parameter.
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # fs: C^1 -> C^1
        t = real_vector_to_complex(ti)
        if time_max is not None:
            self.t_time_max = (
                real_vector_to_complex(
                    torch.ones(
                        t.shape,
                        dtype=self.torch_float_datatype,
                        device=torch.device(device),
                    )
                )
                * time_max
            )
            s = torch.matmul(1 / self.t_time_max.view(-1, 1), self.sdt.view(1, -1))
            f0 = (
                torch.exp((self.rdt * t) / time_max)
                * fs(self.rdt / self.t_time_max)
                / 2.0
            )
            fp = torch.reshape(fs(torch.flatten(s)), s.shape)
            pans = (
                torch.exp((self.rdt * t) / time_max).view(-1, 1)
                * fp
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 1) + f0) * (1 / time_max)
            return result.real
        else:
            s = torch.matmul(1 / t.view(-1, 1), self.sdt.view(1, -1))
            f0 = torch.exp(self.rdt) * fs(self.rdt / t) / 2.0
            fp = torch.reshape(fs(torch.flatten(s)), s.shape)
            pans = (
                torch.exp(self.sdt)
                * fp
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 1) + f0) * (1 / t)
            return result.real

    def compute_s(self, ti, time_max=None):
        r"""Computes :math:`\mathbf{s}` to evaluate the Laplace representation :math:`\mathbf{F}(\mathbf{s})` at, from the input time points :math:`t`, using the selected ILT algorithm.

        Args:
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.

        Returns:
            Tensor of complex s points :math:`\mathbf{s}` of shape :math:`(\text{SeqLen}, \text{ReconTerms})`. :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`.

        """
        # Split of forward into generating s values, then taking function evaluations at s points and integrating along that line
        t = real_vector_to_complex(ti)
        if time_max is not None:
            self.t_time_max = (
                real_vector_to_complex(
                    torch.ones(
                        t.shape,
                        dtype=self.torch_float_datatype,
                        device=torch.device(device),
                    )
                )
                * time_max
            )
            s = torch.matmul(1 / self.t_time_max.view(-1, 1), self.sdt.view(1, -1))
            s0 = self.rdt / self.t_time_max
            return torch.hstack((s0.view(-1, 1), s)), self.rdt / time_max
        else:
            s = torch.matmul(1 / t.view(-1, 1), self.sdt.view(1, -1))
            s0 = self.rdt / t
            return torch.hstack((s0.view(-1, 1), s)), self.rdt / t

    def line_integrate(self, fp, ti, time_max=None):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        t = real_vector_to_complex(ti)
        if time_max is not None:
            f0 = torch.exp((self.rdt * t) / time_max) * fp[:, 0] / 2.0
            pans = (
                torch.exp((self.rdt * t) / time_max).view(-1, 1)
                * fp[:, 1:]
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 1) + f0) * (1 / time_max)
            return result.real
        else:
            f0 = torch.exp(self.rdt) * fp[:, 0] / 2.0
            pans = (
                torch.exp(self.sdt)
                * fp[:, 1:]
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 1) + f0) / t
            return result.real

    def line_integrate_multi(self, fp, ti, time_max=None):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time points
            :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        t = real_vector_to_complex(ti)
        if time_max is not None:
            f0 = torch.exp((self.rdt * t) / time_max) * fp[:, :, 0] / 2.0
            pans = (
                torch.exp((self.rdt * t) / time_max).view(-1, 1)
                * fp[:, :, 1:]
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 2) + f0) * (1 / time_max)
            return result.real
        else:
            f0 = torch.exp(self.rdt) * fp[:, :, 0] / 2.0
            pans = (
                torch.exp(self.sdt)
                * fp[:, :, 1:]
                * (
                    1.0
                    + 1j * self.theta * (1.0 + (1 / torch.tan(self.theta) ** 2))
                    - 1j * (1 / torch.tan(self.theta))
                )
            )
            result = (2.0 / 5.0) * (torch.sum(pans, 2) + f0) / t
            return result.real


class Stehfest(InverseLaplaceTransformAlgorithmBase):
    r"""Inherits from :meth:`torchlaplace.inverse_laplace.InverseLaplaceTransformAlgorithmBase`.

    Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.
    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`.

    Uses a discrete version of the Post-Widder formula that is an approximation for ILT Equation using a power series expansion of real part of :math:`\mathbf{s}`.
    It has internal terms that alternate in sign and become large as the order of approximation is increased, and suffers from numerical precision issues for large orders of approximation.
    It is fairly easy to implement.

    .. note::
        Reference: Al-Shuaibi, A. Inversion of the laplace transform via post—widder formula. Integral Transforms and Special Functions, 11(3):225–232, 2001.

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        # Also computes each s for each time evaluated at
        super().__init__(
            ilt_reconstruction_terms, torch_float_datatype, torch_complex_datatype
        )
        self.torch_complex_datatype = torch_complex_datatype
        M = int((ilt_reconstruction_terms - 1) / 2)
        self.M = M
        self.ln2 = torch.log(torch.Tensor([2.0]).to(device))
        # _coeff routine requires even degree
        if M % 2 > 0:
            self.M += 1

        self.V = self._coeff()
        p_real_dt = (
            torch.arange(
                1, M + 1, dtype=torch_float_datatype, device=torch.device(device)
            )
            * self.ln2
        )
        # NB: s is real
        self.sdt = real_vector_to_complex(p_real_dt)

    def _coeff(self):
        r"""Salzer summation weights (aka, "Stehfest coefficients")
        only depend on the approximation order (M) and the precision"""

        M2 = int(self.M / 2.0)  # checked earlier that M is even

        V = torch.empty(
            (self.M,), dtype=self.torch_complex_datatype, device=torch.device(device)
        )

        def fac(x):
            return float(factorial(x, exact=True))

        # Salzer summation weights
        # get very large in magnitude and oscillate in sign,
        # if the precision is not high enough, there will be
        # catastrophic cancellation
        for k in range(1, self.M + 1):
            z = torch.zeros(
                (min(k, M2) + 1,),
                dtype=self.torch_complex_datatype,
                device=torch.device(device),
            )
            for j in range(int((k + 1) / 2.0), min(k, M2) + 1):
                z[j] = (
                    j**M2
                    * fac(2 * j)
                    / (fac(M2 - j) * fac(j) * fac(j - 1) * fac(k - j) * fac(2 * j - k))
                )
            V[k - 1] = (-1) ** (k + M2) * torch.sum(z)

        return V

    def compute_s(self, ti):
        t = real_vector_to_complex(ti)
        return torch.matmul((1 / t.view(-1, 1)), self.sdt.view(1, -1))

    def line_integrate(self, fp, ti):
        t = real_vector_to_complex(ti)
        result = torch.matmul(fp, self.V) * self.ln2 / t
        # ignore any small imaginary part
        return result.real

    def forward(self, fs, ti):
        # fs: C^1 -> C^1
        t = real_vector_to_complex(ti)
        s = torch.matmul((1 / t.view(-1, 1)), self.sdt.view(1, -1))
        fp = torch.reshape(fs(torch.flatten(s)), s.shape)
        result = torch.matmul(fp, self.V) * self.ln2 / t
        # ignore any small imaginary part
        return result.real


class DeHoog(InverseLaplaceTransformAlgorithmBase):
    r"""Inherits from :meth:`torchlaplace.inverse_laplace.InverseLaplaceTransformAlgorithmBase`.

    Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.
    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`.

    `De Hoog` Is an accelerated version of the Fouier ILT, defined in :meth:`torchlaplace.inverse_laplace.Fourier`.
    It uses a non-linear double acceleration, using Padé approximation along with a remainder term for the series.
    This is somewhat complicated to implement, due to the recurrence operations to represent the Padé approximation, due to this although higher precision, the gradients
    have to propagate through many recurrence relation paths, making it slow to use in practice compared to Fourier (FSI), however more accurate when we can afford the additional time complexity.

    .. note::
        Reference: De Hoog, F. R., Knight, J., and Stokes, A. An improved method for numerical inversion of laplace transforms.
        SIAM Journal on Scientific and Statistical Computing, 3(3):357–366, 1982

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
        alpha (float): :math:`\alpha`, default :math:`\alpha=10e-10`.
        tol (float): desired tolerance, if not specified simply related to alpha as :math:`\text{tol}=10\alpha`.
        scale (float): scaling factor (tuneable), default 2.0.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        alpha=1.0e-10,
        tol=None,
        scale=2.0,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        super().__init__(
            ilt_reconstruction_terms, torch_float_datatype, torch_complex_datatype
        )
        M = int((ilt_reconstruction_terms - 1) / 2)
        self.M = M
        self.alpha = torch.Tensor([alpha]).to(device)
        # desired tolerance (here simply related to alpha)
        if tol is not None:
            self.tol = torch.Tensor([tol]).to(device)
        else:
            self.tol = self.alpha * 10.0
        self.nt = ilt_reconstruction_terms  # number of terms in approximation
        # scaling factor (likely tune-able, but 2 is typical)
        self.scale = torch.Tensor([scale]).to(device)
        self.torch_float_datatype = torch_float_datatype
        self.torch_complex_datatype = torch_complex_datatype

    def compute_fixed_s(self, time_max):
        r"""Computes :math:`\mathbf{s}` to evaluate the Laplace representation :math:`\mathbf{F}(\mathbf{s})` at for a single input time point `time_max`, using the selected ILT algorithm.

        Args:
            time_max (float): Maximum time to compute reconstruction up to, a single time point to generate `s` for.

        Returns:
            Tensor of complex s points :math:`\mathbf{s}` of shape :math:`(\text{ReconTerms})`.

        """
        NT = self.nt
        tmax = torch.Tensor([time_max]).to(device)
        T = torch.Tensor([self.scale * tmax]).to(device)

        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        s = (
            gamma
            + 1j
            * torch.pi
            * torch.arange(
                NT, dtype=self.torch_float_datatype, device=torch.device(device)
            )
            / T
        )
        return s

    def fixed_line_integrate(self, fp, ti, time_max):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for most ILT
            algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.
            time_max (float): Maximum time to compute reconstruction up to, a single time point to generate `s` for.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time points
            :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # fs: C^1 -> C^1
        tv = real_vector_to_complex(ti)
        M = self.M
        NT = self.nt
        tmax = torch.Tensor([time_max]).to(device)
        T = torch.Tensor([self.scale * tmax]).to(device)
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)

        # would it be useful to try re-using
        # space between e&q ?
        e = torch.empty(
            (NT, M + 1), dtype=self.torch_complex_datatype, device=torch.device(device)
        )
        q = torch.empty(
            (2 * M, M), dtype=self.torch_complex_datatype, device=torch.device(device)
        )
        d = torch.empty(
            (NT,), dtype=self.torch_complex_datatype, device=torch.device(device)
        )

        es = []
        qs = []

        # initialize Q-D table
        q = fp[1 : 2 * M + 1] / fp[0 : 2 * M]
        q[0] = fp[1] / (fp[0] / 2.0)
        qs.append(q)

        # rhombus rule for filling triangular Q-D table (e & q)
        for r in range(1, M + 1):
            # start with e, column 1, 0:2*M-2
            mr = 2 * (M - r) + 1
            if r == 1:
                e = torch.zeros(
                    (NT,),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[0:mr] = qs[r - 1][1 : mr + 1] - qs[r - 1][0:mr]
                es.append(e)
            else:
                e = torch.zeros(
                    (NT,),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[0:mr] = (
                    qs[r - 1][1 : mr + 1] - qs[r - 1][0:mr] + es[r - 2][1 : mr + 1]
                )
                es.append(e)
            if not r == M:
                q = torch.zeros(
                    (2 * M,),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                q[0:mr] = (
                    qs[r - 1][1 : mr + 1] * es[r - 1][1 : mr + 1] / es[r - 1][0:mr]
                )
                qs.append(q)

        # build up continued fraction coefficients (d)
        d[0] = fp[0] / 2.0
        for r in range(1, M + 1):
            d[2 * r - 1] = -qs[r - 1][0]  # even terms
            d[2 * r] = -es[r - 1][0]  # odd terms

        # seed A and B for recurrence
        Aim1 = torch.view_as_complex(torch.Tensor([0, 0]).to(device))
        Ai = d[0]
        Bim1 = torch.view_as_complex(torch.Tensor([1, 0]).to(device))
        Bi = torch.view_as_complex(torch.Tensor([1, 0]).to(device))

        # base of the power series
        z = torch.exp(1j * torch.pi * ti / T)

        # coefficients of Pade approximation (A & B)
        # using recurrence for all but last term
        for i in range(1, 2 * M):
            Ait = Ai + d[i] * Aim1 * z
            Aim1 = Ai
            Ai = Ait
            Bit = Bi + d[i] * Bim1 * z
            Bim1 = Bi
            Bi = Bit

        # "improved remainder" to continued fraction
        brem = (1.0 + (d[2 * M - 1] - d[2 * M]) * z) / 2.0
        rem = -brem * (1.0 - torch.sqrt(1.0 + d[2 * M] * z / brem**2))

        # last term of recurrence using new remainder
        A_2mp1 = Ai + rem * Aim1
        B_2mp1 = Bi + rem * Bim1

        # diagonal Pade approximation
        # F=A/B represents accelerated trapezoid rule
        return (torch.exp(gamma * tv) / T * (A_2mp1 / B_2mp1).real).real

    def compute_s(self, ti, time_max=None, Ti=None):
        r"""Computes :math:`\mathbf{s}` to evaluate the Laplace representation :math:`\mathbf{F}(\mathbf{s})` at, from the input time points :math:`t`, using the selected ILT algorithm.

        Args:
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.
            Ti (float): Scaled maximum time to compute reconstruction up to. Best results use default of None, as this uses `T=time_max * self.scale`.

        Returns:
            Tensor of complex s points :math:`\mathbf{s}` of shape :math:`(\text{SeqLen}, \text{ReconTerms})`. :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`.

        """
        tv = real_vector_to_complex(ti)
        NT = self.nt
        if time_max is not None:
            tmax = torch.Tensor([time_max]).to(device) * torch.ones_like(tv)
        else:
            tmax = tv
        if Ti is not None:
            assert type(Ti) == self.torch_float_datatype
            T = Ti
        else:
            T = tmax * self.scale
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        si = torch.matmul(
            1 / T.view(-1, 1),
            1j
            * torch.pi
            * torch.arange(
                NT, dtype=self.torch_float_datatype, device=torch.device(device)
            ).view(1, -1),
        )
        return gamma.view(-1, 1) + si, T

    def line_integrate(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for most ILT
            algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time points
            :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        t = real_vector_to_complex(ti)
        M = self.M
        NT = self.nt
        t_samples = t.shape[0]
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)

        # would it be useful to try re-using
        # space between e&q ?
        e = torch.empty(
            (t_samples, NT, M + 1),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )
        q = torch.empty(
            (t_samples, 2 * M, M),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )
        d = torch.empty(
            (
                t_samples,
                NT,
            ),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )

        es = []
        qs = []

        # initialize Q-D table
        q = fp[:, 1 : 2 * M + 1] / fp[:, 0 : 2 * M]
        q[:, 0] = fp[:, 1] / (fp[:, 0] / 2.0)
        qs.append(q)

        # rhombus rule for filling triangular Q-D table (e & q)
        for r in range(1, M + 1):
            # start with e, column 1, 0:2*M-2
            mr = 2 * (M - r) + 1
            if r == 1:
                e = torch.zeros(
                    (
                        t_samples,
                        NT,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[:, 0:mr] = qs[r - 1][:, 1 : mr + 1] - qs[r - 1][:, 0:mr]
                es.append(e)
            else:
                e = torch.zeros(
                    (
                        t_samples,
                        NT,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[:, 0:mr] = (
                    qs[r - 1][:, 1 : mr + 1]
                    - qs[r - 1][:, 0:mr]
                    + es[r - 2][:, 1 : mr + 1]
                )
                es.append(e)
            if not r == M:
                q = torch.zeros(
                    (
                        t_samples,
                        2 * M,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                q[:, 0:mr] = (
                    qs[r - 1][:, 1 : mr + 1]
                    * es[r - 1][:, 1 : mr + 1]
                    / es[r - 1][:, 0:mr]
                )
                qs.append(q)

        # build up continued fraction coefficients (d)
        d[:, 0] = fp[:, 0] / 2.0
        for r in range(1, M + 1):
            d[:, 2 * r - 1] = -qs[r - 1][:, 0]  # even terms
            d[:, 2 * r] = -es[r - 1][:, 0]  # odd terms

        # seed A and B for recurrence
        Aim1 = torch.view_as_complex(torch.Tensor([0, 0]).to(device))
        Ai = d[:, 0]
        Bim1 = torch.view_as_complex(torch.Tensor([1, 0]).to(device))
        Bi = torch.view_as_complex(torch.Tensor([1, 0]).to(device))

        # base of the power series
        z = torch.exp(1j * torch.pi * t / T)

        # coefficients of Pade approximation (A & B)
        # using recurrence for all but last term
        for i in range(1, 2 * M):
            Ait = Ai + d[:, i] * Aim1 * z
            Aim1 = Ai
            Ai = Ait
            Bit = Bi + d[:, i] * Bim1 * z
            Bim1 = Bi
            Bi = Bit

        # "improved remainder" to continued fraction
        brem = (1.0 + (d[:, 2 * M - 1] - d[:, 2 * M]) * z) / 2.0
        rem = -brem * (1.0 - torch.sqrt(1.0 + d[:, 2 * M] * z / brem**2))

        # last term of recurrence using new remainder
        A_2mp1 = Ai + rem * Aim1
        B_2mp1 = Bi + rem * Bim1

        # diagonal Pade approximation
        # F=A/B represents accelerated trapezoid rule
        result = torch.exp(gamma * t) / T * (A_2mp1 / B_2mp1).real

        return result.real

    def line_integrate_all_multi(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm (takes batch input of `fp`).

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for most ILT
            algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different
            time points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # Input shapes
        # fp [batch, times, data_dims, approximation_terms * 2 + 1]
        # ti [times]
        # T [times]
        # Returns
        # [batch, times, data_dims]
        samples = []
        for traj in fp:
            data_dims = []
            for data_dim_idx in range(fp.shape[2]):
                data_dims.append(self.line_integrate(traj[:, data_dim_idx, :], ti, T))
            samples.append(torch.stack(data_dims, 1))
        return torch.stack(samples)

    def forward(self, fs, ti, time_max=None, Ti=None):
        r"""Reconstructs a trajectory :math:`\mathbf{x}(t)` for a Laplace representation :math:`\mathbf{F}(\mathbf{s})`, at time points :math:`t`.

        Args:
            fs (Torch.nn.Module or Callable): The first parameter.
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.
            Ti (float): Scaled maximum time to compute reconstruction up to. Best results use default of None, as this uses `T=time_max * self.scale`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time points
            :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # fs: C^1 -> C^1
        t = real_vector_to_complex(ti)
        M = self.M
        NT = self.nt
        t_samples = t.shape[0]

        if time_max is not None:
            tmax = torch.Tensor([time_max]).to(device) * torch.ones_like(t)
        else:
            tmax = t
        if Ti is not None:
            assert type(Ti) == self.torch_float_datatype
            T = Ti
        else:
            T = tmax * self.scale
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        si = torch.matmul(
            1 / T.view(-1, 1),
            1j
            * torch.pi
            * torch.arange(
                NT, dtype=self.torch_float_datatype, device=torch.device(device)
            ).view(1, -1),
        )
        fp = fs(gamma.view(-1, 1) + si)

        # would it be useful to try re-using
        # space between e&q ?
        e = torch.empty(
            (t_samples, NT, M + 1),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )
        q = torch.empty(
            (t_samples, 2 * M, M),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )
        d = torch.empty(
            (
                t_samples,
                NT,
            ),
            dtype=self.torch_complex_datatype,
            device=torch.device(device),
        )

        es = []
        qs = []

        # initialize Q-D table
        q = fp[:, 1 : 2 * M + 1] / fp[:, 0 : 2 * M]
        q[:, 0] = fp[:, 1] / (fp[:, 0] / 2.0)
        qs.append(q)

        # rhombus rule for filling triangular Q-D table (e & q)
        for r in range(1, M + 1):
            # start with e, column 1, 0:2*M-2
            mr = 2 * (M - r) + 1
            if r == 1:
                e = torch.zeros(
                    (
                        t_samples,
                        NT,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[:, 0:mr] = qs[r - 1][:, 1 : mr + 1] - qs[r - 1][:, 0:mr]
                es.append(e)
            else:
                e = torch.zeros(
                    (
                        t_samples,
                        NT,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                e[:, 0:mr] = (
                    qs[r - 1][:, 1 : mr + 1]
                    - qs[r - 1][:, 0:mr]
                    + es[r - 2][:, 1 : mr + 1]
                )
                es.append(e)
            if not r == M:
                q = torch.zeros(
                    (
                        t_samples,
                        2 * M,
                    ),
                    dtype=self.torch_complex_datatype,
                    device=torch.device(device),
                )
                q[:, 0:mr] = (
                    qs[r - 1][:, 1 : mr + 1]
                    * es[r - 1][:, 1 : mr + 1]
                    / es[r - 1][:, 0:mr]
                )
                qs.append(q)

        # build up continued fraction coefficients (d)
        d[:, 0] = fp[:, 0] / 2.0
        for r in range(1, M + 1):
            d[:, 2 * r - 1] = -qs[r - 1][:, 0]  # even terms
            d[:, 2 * r] = -es[r - 1][:, 0]  # odd terms

        # seed A and B for recurrence
        Aim1 = torch.view_as_complex(torch.Tensor([0, 0]).to(device))
        Ai = d[:, 0]
        Bim1 = torch.view_as_complex(torch.Tensor([1, 0]).to(device))
        Bi = torch.view_as_complex(torch.Tensor([1, 0]).to(device))

        # base of the power series
        z = torch.exp(1j * torch.pi * t / T)

        # coefficients of Pade approximation (A & B)
        # using recurrence for all but last term
        for i in range(1, 2 * M):
            Ait = Ai + d[:, i] * Aim1 * z
            Aim1 = Ai
            Ai = Ait
            Bit = Bi + d[:, i] * Bim1 * z
            Bim1 = Bi
            Bi = Bit

        # "improved remainder" to continued fraction
        brem = (1.0 + (d[:, 2 * M - 1] - d[:, 2 * M]) * z) / 2.0
        rem = -brem * (1.0 - torch.sqrt(1.0 + d[:, 2 * M] * z / brem**2))

        # last term of recurrence using new remainder
        A_2mp1 = Ai + rem * Aim1
        B_2mp1 = Bi + rem * Bim1

        # diagonal Pade approximation
        # F=A/B represents accelerated trapezoid rule
        result = (torch.exp(gamma * t) / T) * (A_2mp1 / B_2mp1).real
        return result.real


class Fourier(InverseLaplaceTransformAlgorithmBase):
    r"""Inherits from :meth:`torchlaplace.inverse_laplace.InverseLaplaceTransformAlgorithmBase`.

    Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.
    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`.

    Expands ILT Equation into an expanded Fourier transform, approximating it with the trapezoidal rule.
    This keeps the Bromwich contour parallel to the imaginary axis, and shifts it along the real axis, i.e. :math:`\sigma \propto \frac{1}{t}`.
    It is fairly easy to implement and scale to multiple dimensions. We denote :math:`s=\sigma + i \omega` and we can express it as,

    .. math::
        \begin{split}
            \mathbf{x}(t) & =  \frac{1}{\pi}e^{\sigma t} \int_0^\infty \Re \left\{ F(s) e^{i \omega t} \right\} d \omega \\
            & \approx \frac{1}{T} e^{\sigma t} \left[ \frac{F(\sigma)}{2}  + \sum_{k=1}^{2N} \Re \left\{  F \left( \sigma + \frac{ik\pi}{T} \right)e^{\frac{ik \pi t}{T}} \right\}      \right]
        \end{split}

    Where we approximate the first Fourier ILT, Equation as a discretized version, using the trapezoidal rule with step size :math:`\frac{\pi}{T}` and evaluating :math:`s` at the
    approximation points :math:`s_k=\sigma + \frac{ik\pi}{T}` in the trapezoidal summation.
    We set the parameters of :math:`\sigma=\alpha-\frac{\log(\epsilon)}{T}`, with :math:`\alpha=1e-3`, :math:`\epsilon=10\alpha`, and the scaling parameter :math:`T=2t`. This gives the query function,

    .. math::
        \begin{split}
            s_k(t) & = \text{1e-3} - \frac{\log(\text{1e-2})}{2t} + \frac{ik\pi}{2t} \\
            \mathcal{Q}(t) & = [s_0(t), \ldots, s_{2N}(t)]^T
        \end{split}

    Where we model the equation with :math:`2N + 1` reconstruction terms, setting :math:`N=16` in experiments, and use double point floating precision to increase the numerical precision of the ILT.

    The ILT-FSI equation provides guarantees that we can always find the inverse from time :math:`t: 0 \rightarrow \infty`, given that the singularities of the system (i.e. the
    points at which :math:`F(s) \to \infty`) lie left of the contour of integration, and this puts no constraint on the imaginary frequency components we can model.
    Of course in practice, we often do not model time at :math:`\infty` and instead model up to a fixed time in the future, which then bounds the exponentially increasing system
    trajectories, and their associated system poles that we can model :math:`\sigma \propto \frac{1}{t}`.

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
        alpha (float): :math:`\alpha`, default :math:`\alpha=1e-3`.
        tol (float): desired tolerance, if not specified simply related to alpha as :math:`\text{tol}=10\alpha`.
        scale (float): scaling factor (tuneable), default 2.0.
        eps (float): Small machine floating point precision, default :math:`\text{eps}=1e-6`.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        alpha=1.0e-3,
        tol=None,
        scale=2.0,
        eps=1e-6,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        # Unaccelerated Fourier Estimate of DeHoog (Ref: F. R. DE HOOG, J. H. KNIGHT AND A. N. STOKES, p. 360)
        super().__init__(
            ilt_reconstruction_terms, torch_float_datatype, torch_complex_datatype
        )
        M = int((ilt_reconstruction_terms - 1) / 2)
        self.M = M
        self.alpha = torch.Tensor([alpha]).to(device)
        # desired tolerance (here simply related to alpha)
        if tol is not None:
            self.tol = torch.Tensor([tol]).to(device)
        else:
            self.tol = self.alpha * 10.0
        self.nt = ilt_reconstruction_terms  # number of terms in approximation
        # scaling factor (likely tune-able, but 2 is typical)
        self.scale = torch.Tensor([scale]).to(device)
        self.k = torch.arange(
            self.nt, dtype=torch_float_datatype, device=torch.device(device)
        )
        self.eps = eps
        self.torch_float_datatype = torch_float_datatype

    def compute_s(self, ti, time_max=None, Ti=None):
        r"""Computes :math:`\mathbf{s}` to evaluate the Laplace representation :math:`\mathbf{F}(\mathbf{s})` at, from the input time points :math:`t`, using the selected ILT algorithm.

        Args:
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.
            Ti (float): Scaled maximum time to compute reconstruction up to. Best results use default of None, as this uses `T=time_max * self.scale`.

        Returns:
            Tensor of complex s points :math:`\mathbf{s}` of shape :math:`(\text{SeqLen}, \text{ReconTerms})`. :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`.

        """
        t = real_vector_to_complex(ti)
        if time_max is not None:
            tmax = torch.Tensor([time_max]).to(device) * torch.ones_like(t)
        else:
            tmax = t  # + self.eps
        if Ti is not None:
            assert type(Ti) == self.torch_float_datatype
            T = Ti
        else:
            T = tmax * self.scale
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        si = torch.matmul(1 / T.view(-1, 1), 1j * torch.pi * self.k.view(1, -1))
        return gamma.view(-1, 1) + si, T

    def line_integrate(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for
            most ILT algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        t = real_vector_to_complex(ti)
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        return (
            (1 / T)
            * torch.exp(gamma * t)
            * (
                fp[:, 0] / 2.0
                + torch.sum(
                    fp[:, 1:]
                    * torch.exp(
                        torch.matmul(
                            (t / T).view(-1, 1), 1j * self.k[1:].view(1, -1) * torch.pi
                        )
                    ),
                    1,
                )
            )
        ).real

    def line_integrate_all_multi(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm (takes batch input of `fp`).

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for most ILT algorithms is
            to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{BatchSize}, \text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # Input shapes
        # fp [batch, times, data_dims, approximation_terms * 2 + 1]
        # ti [times]
        # T [times]
        # Returns
        # [batch, times, data_dims]
        t = real_vector_to_complex(ti)  # Could be slowing down ?
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        return (
            ((1 / T) * torch.exp(gamma * t)).view(1, -1, 1)
            * (
                fp[:, :, :, 0] / 2.0
                + torch.sum(
                    fp[:, :, :, 1:]
                    * torch.exp(
                        torch.matmul(
                            (t / T).view(-1, 1), 1j * self.k[1:].view(1, -1) * torch.pi
                        )
                    ).view(fp.shape[1], 1, fp.shape[3] - 1),
                    3,
                )
            )
        ).real

    def line_integrate_multi(self, fp, ti, T):
        r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for `fp`, Laplace representations evaluated at `s` points from the input `ti` points, :math:`t`, using the selected ILT algorithm.

        Args:
            fp (Tensor): Laplace representation evaluated at `s` points derived from the input time points `ti`. `fp` has shape :math:`(\text{SeqLen}, d_{\text{obs}}, \text{ReconTerms})`
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            T (Tensor): time points to reconstruct trajectory for used as the reconstruction of times up to `T` time point of shape :math:`(\text{SeqLen})`. Best practice for
            most ILT algorithms is to set `T = ti`, for the ILT algorithms that rely on `T`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time
            points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        t = real_vector_to_complex(ti)
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        return (
            (1 / T)
            * torch.exp(gamma * t)
            * (
                fp[:, :, 0] / 2.0
                + torch.sum(
                    fp[:, :, 1:]
                    * torch.exp(
                        torch.matmul(
                            (t / T).view(-1, 1), 1j * self.k[1:].view(1, -1) * torch.pi
                        )
                    ),
                    2,
                )
            )
        ).real

    def forward(self, fs, ti, time_max=None, Ti=None):
        r"""Reconstructs a trajectory :math:`\mathbf{x}(t)` for a Laplace representation :math:`\mathbf{F}(\mathbf{s})`, at time points :math:`t`.

        Args:
            fs (Torch.nn.Module or Callable): The first parameter.
            ti (Tensor): time points to reconstruct trajectory for of shape :math:`(\text{SeqLen})`.
            time_max (float): Maximum time to compute reconstruction up to. Best results use default of None, as this uses `time_max=ti`.
            Ti (float): Scaled maximum time to compute reconstruction up to. Best results use default of None, as this uses `T=time_max * self.scale`.

        Returns:
            Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different
            time points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

        """
        # fs: C^1 -> C^1
        t = real_vector_to_complex(ti)
        if time_max is not None:
            tmax = torch.Tensor([time_max]).to(device) * torch.ones_like(t)
        else:
            tmax = t
        if Ti is not None:
            assert type(Ti) == self.torch_float_datatype
            T = Ti
        else:
            T = tmax * self.scale
        gamma = self.alpha - torch.log(self.tol) / (self.scale * T)
        si = torch.matmul(1 / T.view(-1, 1), 1j * torch.pi * self.k.view(1, -1))
        fp = fs(gamma.view(-1, 1) + si)
        return (
            (1 / T)
            * torch.exp(gamma * t)
            * (
                fp[:, 0] / 2.0
                + torch.sum(
                    fp[:, 1:]
                    * torch.exp(
                        torch.matmul(
                            (t / T).view(-1, 1), 1j * self.k[1:].view(1, -1) * torch.pi
                        )
                    ),
                    1,
                )
            )
        ).real


class CME(InverseLaplaceTransformAlgorithmBase):
    r"""Inherits from :meth:`torchlaplace.inverse_laplace.InverseLaplaceTransformAlgorithmBase`.

    Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.
    Given a parameterized Laplace representation function :math:`\mathbf{F}(\mathbf{s})`.

    Concentrated matrix exponential (CME), uses a similar form to that of the Fourier Series Inverse, approximating ILT Equation with the trapezoidal rule. This uses the form of,

    .. math::
        \begin{aligned}
            \mathbf{x}(t) \approx \frac{1}{T}\sum_{k=1}^{2N} \eta_k F \left(\frac{\beta_k}{T}\right)
        \end{aligned}

    The coefficients :math:`\eta_k, \beta_k` are determined by a complex procedure, with a numerical optimization step involved.
    This provides a good approximation for the reconstruction and the coefficients of up to a pre-specified order can be pre-computed and cached for low complexity run time.
    Similarly to Fourier (FSI), CMEs Bromwich contour remains parallel to the imaginary axis and is shifted along the real axis, i.e. :math:`\sigma \propto \frac{1}{t}`.
    It is moderately easy to implement when using pre-computed coefficients and scale to multiple dimensions.

    .. note::
        Reference: Horváth, G., Horváth, I., Almousa, S. A.-D., and Telek, M. Numerical inverse laplace transformation using concentrated matrix exponential distributions. Performance Evaluation, 137:102067, 2020.

    Args:
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `fs` to reconstruct a single time point.
        torch_float_datatype (Torch.dtype): Torch float datatype to use internally in the ILT algorithm, and also output data type of :math:`\mathbf{x}(t)`. Default `torch.float32`.
        torch_complex_datatype (Torch.dtype): Torch complex datatype to use internally in the ILT algorithm. Default `torch.cfloat`.
    """

    def __init__(
        self,
        ilt_reconstruction_terms=33,
        torch_float_datatype=TORCH_FLOAT_DATATYPE,
        torch_complex_datatype=TORCH_COMPLEX_DATATYPE,
    ):
        super().__init__(
            ilt_reconstruction_terms, torch_float_datatype, torch_complex_datatype
        )
        M = int((ilt_reconstruction_terms - 1) / 2)
        self.M = M
        cme_params = cme_params_factory()
        params = cme_params[0]
        max_fn_evals = ilt_reconstruction_terms
        for p in cme_params:
            if p["cv2"] < params["cv2"] and p["n"] + 1 <= max_fn_evals:
                params = p
        eta = (
            np.concatenate(
                ([params["c"]], np.array(params["a"]) + 1j * np.array(params["b"]))
            )
            * params["mu1"]
        )
        beta = (
            np.concatenate(
                ([1], 1 + 1j * np.arange(1, params["n"] + 1) * params["omega"])
            )
            * params["mu1"]
        )
        self.eta = complex_numpy_to_complex_torch(eta).to(device)
        self.beta = complex_numpy_to_complex_torch(beta).to(device)

    def compute_s(self, ti):
        t = real_vector_to_complex(ti)
        s = torch.matmul((1 / t).view(-1, 1), self.beta.view(1, -1))
        return s, t

    def line_integrate(self, fp, ti):
        # t = real_vector_to_complex(ti)
        return (torch.matmul(fp, self.eta.view(-1, 1)).view(-1) / ti).real

    def line_integrate_all_multi(self, fp, ti, T):
        # Input shapes
        # fp [batch, times, data_dims, approximation_terms * 2 + 1]
        # ti [times]
        # T [times]
        # Returns
        # [batch, times, data_dims]
        # t = real_vector_to_complex(ti)
        return (
            (1 / T).view(1, -1, 1) * torch.matmul(fp, self.eta.view(-1, 1)).squeeze(-1)
        ).real

    def forward(self, fs, ti):
        # fs: C^1 -> C^1
        t = real_vector_to_complex(ti)
        s = torch.matmul((1 / t).view(-1, 1), self.beta.view(1, -1))
        return (torch.matmul(fs(s), self.eta.view(-1, 1)).view(-1) / t).real


if __name__ == "__main__":
    ILT_RECONSTRUCTION_TERMS = 33

    t = torch.linspace(0.0001, 10.0, 1000).to(device)

    # Exponential
    # sign = +1  # (-1 for increasing exponential, +1 for decaying exponential)
    # def fs(so): return 1 / (so + 1 * sign)  # Laplace solution
    # def ft(t): return torch.exp(-t * sign)  # Time solution

    # # Cosine
    def fs(so):
        return so / (so**2 + 1)  # Laplace solution

    def ft(t):
        return torch.cos(t)  # Time solution

    print("")

    # Tablot

    # Evaluate s points per time input (Default, as more accurate inversion)

    decoder = FixedTablot(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t)
    print(
        "FixedTablot Loss:\t{}\t\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Split evaluation of s points out from that of the line integral (should be the exact same result as above)
    decoder = FixedTablot(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s, _ = decoder.compute_s(t)
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t)
    print(
        "FixedTablot Loss (Split apart):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    decoder = FixedTablot(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s, _ = decoder.compute_s(t, time_max=torch.max(t))
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t, time_max=t.max().item())
    print(
        "FixedTablot Loss (Split apart, Fixed Max Time):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Evaluate s points for one fixed time, maximum time (Less accurate, maybe more stable ?)

    decoder = FixedTablot(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t, time_max=torch.max(t))
    print(
        "FixedTablot Loss (Fixed Max Time):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Stehfest - Increasing degree here, introduces numerical error that increases larger than other methods, therefore for high degree becomes unstable

    decoder = Stehfest(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t)
    print(
        "Stehfest Loss:\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    decoder = Stehfest(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s = decoder.compute_s(t)
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t)
    print(
        "Stehfest Loss (Split apart):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Fourier (Un accelerated DeHoog)
    decoder = Fourier(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t)
    print(
        "Fourier (Un accelerated DeHoog) Loss:\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    decoder = Fourier(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s, T = decoder.compute_s(t)
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t, T)
    print(
        "Fourier (Un accelerated DeHoog) Loss (Split apart):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # DeHoog

    decoder = DeHoog(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t)
    print(
        "DeHoog Loss:\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Split evaluation of s points out from that of the line integral (should be the exact same result as above)
    decoder = DeHoog(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s, T = decoder.compute_s(t)
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t, T)
    print(
        "DeHoog Loss (Split apart):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # Single line integral
    decoder = DeHoog(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s = decoder.compute_fixed_s(torch.max(t))
    fh = fs(s)
    f_hat_t = decoder.fixed_line_integrate(fh, t, torch.max(t))
    print(
        "DeHoog Loss (Fixed Line Integrate):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    decoder = DeHoog(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t, time_max=torch.max(t))
    print(
        "DeHoog Loss (Fixed Max Time):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    # CME
    decoder = CME(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    f_hat_t = decoder(fs, t)
    print(
        "CME Loss:\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )

    decoder = CME(ilt_reconstruction_terms=ILT_RECONSTRUCTION_TERMS).to(device)
    t0 = time()
    s, T = decoder.compute_s(t)
    fh = fs(s)
    f_hat_t = decoder.line_integrate(fh, t, T)
    print(
        "CME Loss (Split apart):\t{}\t| time: {}".format(
            np.sqrt(torch.nn.MSELoss()(ft(t), f_hat_t).cpu().numpy()), time() - t0
        )
    )
    print("Done")



###########################
# Neural Laplace: Learning diverse classes of differential equations in the Laplace domain
# Author: Samuel Holt
###########################
"""
Laplace Reconstructions
"""
import torch
from torch import Tensor, nn

from .inverse_laplace import CME, DeHoog, FixedTablot, Fourier, Stehfest
from .transformations import complex_to_spherical_riemann, spherical_to_complex

ILT_ALGORITHMS = {
    "fourier": Fourier,
    "dehoog": DeHoog,
    "cme": CME,
    "fixed_tablot": FixedTablot,
    "stehfest": Stehfest,
}


def laplace_reconstruct(
    laplace_rep_func,
    p,
    t,
    recon_dim=None,
    ilt_algorithm="fourier",
    use_sphere_projection=True,
    ilt_reconstruction_terms=33,
    options=None,
) -> Tensor:
    r"""Reconstruct trajectories :math:`\mathbf{x}(t)` for a system of Laplace representations.

    Given a parameterized Laplace representation functional :math:`\mathbf{F}(\mathbf{p},\mathbf{s})`,

    .. math::
        \begin{aligned}
            \mathbf{x}(t) & = \text{inverse_laplace_transform}(\mathbf{F}(\mathbf{p},\mathbf{s}), t)
        \end{aligned}

    Where :math:`\mathbf{p}` is a Tensor encoding the initial system state as a latent variable, and :math:`t` is the time points to reconstruct trajectories for.

    The parameterized Laplace representation functional :attr:`laplace_rep_func`, :math:`\mathbf{F}(\mathbf{p},\mathbf{s})` also takes an input complex value :math:`\mathbf{s}`.
    This :math:`\mathbf{s}` is used internally when reconstructing a specified time point with the selected inverse Laplace transform algorithm :attr:`ilt_algorithm`.

    Output dtypes and numerical precision are based on the dtypes of the inputs :attr:`p`.

    The reconstructions :math:`\mathbf{x}(t)` are of shape :math:`(\text{MiniBatchSize}, \text{SeqLen}, d_{\text{obs}})`. Where :math:`\text{SeqLen}` dimension corresponds to the input evaluated time points :math:`t`, and :math:`d_{\text{obs}}` is the trajectory dimension for a given time point, :math:`\mathbf{x}(t)`.

    Args:
        laplace_rep_func (nn.Module): Function that maps an input of two tensors, first a scalar Tensor `p` encoding the initial system state and secondly a complex Tensor `s` used to evaluate the Laplace representation.
        p (Tensor): latent variable Tensor of shape :math:`(\text{MiniBatchSize}, \text{K})`. Where :math:`\text{K}` is a hyperparameter, and can be set by the user to their desired value.
        t (Tensor): time points to reconstruct trajectories for of shape :math:`(\text{MiniBatchSize}, \text{SeqLen})` or :math:`(\text{SeqLen})` if all trajectories use the same time points.
        recon_dim (int): trajectory dimension for a given time point. Corresponds to dim :math:`d_{\text{obs}}`. If not explicitly specified, will use the same last dimension of `p`, i.e. :math:`\text{K}`.
        ilt_algorithm (str): inverse Laplace transform algorithm to use. Default: ``fourier``. Available are {``fourier``, ``dehoog``, ``cme``, ``fixed_tablot``, ``stehfest``}. See inverse_laplace.py for further details.
        use_sphere_projection (bool): this uses the `laplace_rep_func` in the stereographic projection of the Riemann sphere. Default ``True``.
        ilt_reconstruction_terms (int): number of ILT reconstruction terms, i.e. the number of complex :math:`s` points in `laplace_rep_func` to reconstruct a single time point.

    Returns:
        Tensor of reconstructions :math:`\mathbf{x}(t)` of shape :math:`(\text{MiniBatchSize}, \text{SeqLen}, d_{\text{obs}})` or if no mini batch of shape :math:`(\text{SeqLen}, d_{\text{obs}})`. :math:`\text{SeqLen}` dimension corresponds to the different time points :math:`t`. This tensor of reconstructions contains the solved value of :math:`\mathbf{x}` for each desired time point in `t`.

    Raises:
        ValueError: if an invalid `ilt_algorithm` is provided.

    """
    options, recon_dim, p = _check_inputs(
        laplace_rep_func,
        p,
        t,
        recon_dim,
        ilt_algorithm,
        use_sphere_projection,
        ilt_reconstruction_terms,
        options,
    )

    if p.dtype == torch.float64:  # For double precision
        ilt = ILT_ALGORITHMS[ilt_algorithm](
            ilt_reconstruction_terms=ilt_reconstruction_terms,
            torch_float_datatype=torch.double,
            torch_complex_datatype=torch.cdouble,
            **options
        )
    else:
        ilt = ILT_ALGORITHMS[ilt_algorithm](
            ilt_reconstruction_terms=ilt_reconstruction_terms, **options
        )
    t = torch.squeeze(t)
    time_dim = t.shape[0]
    batch_dim = p.shape[0]
    s, T = ilt.compute_s(t)
    T = T
    if use_sphere_projection:
        thetam, phim = complex_to_spherical_riemann(
            torch.flatten(s.real), torch.flatten(s.imag)
        )
        thetam = torch.reshape(thetam, s.real.shape)
        phim = torch.reshape(phim, s.imag.shape)
        sph_coords = torch.cat((thetam, phim), 1)
        s_terms_dim = thetam.shape[1]
        inputs = torch.cat(
            (
                sph_coords.view(1, time_dim, -1).repeat(batch_dim, 1, 1),
                p.view(batch_dim, 1, -1).repeat(1, time_dim, 1),
            ),
            2,
        )
        theta, phi = laplace_rep_func(inputs)
        sr = spherical_to_complex(theta, phi)
        ss = sr.view(-1, time_dim, recon_dim, s_terms_dim)
    else:
        s_split = torch.cat((s.real, s.imag), 1)
        s_terms_dim = s.shape[1]
        inputs = torch.cat(
            (
                s_split.view(1, time_dim, -1).repeat(batch_dim, 1, 1),
                p.view(batch_dim, 1, -1).repeat(1, time_dim, 1),
            ),
            2,
        )
        s_real, s_imag = laplace_rep_func(inputs)
        so = torch.view_as_complex(
            torch.stack((torch.flatten(s_real), torch.flatten(s_imag)), 1)
        )
        sr = torch.reshape(so, s_real.shape)
        ss = sr.view(-1, time_dim, recon_dim, s_terms_dim)
    return ilt.line_integrate_all_multi(ss, t, T)


def _check_inputs(
    laplace_rep_func,
    p,
    t,
    recon_dim,
    ilt_algorithm,
    use_sphere_projection,
    ilt_reconstruction_terms,
    options,
):
    assert isinstance(
        laplace_rep_func, nn.Module
    ), "laplace_rep_func must be a descendant of torch.nn.Module"
    assert isinstance(p, Tensor), "p must be a torch.Tensor type"
    assert isinstance(t, Tensor), "t must be a torch.Tensor type"
    assert isinstance(
        use_sphere_projection, bool
    ), "use_sphere_projection must be a bool type"
    if ilt_algorithm not in ILT_ALGORITHMS:
        raise ValueError(
            'Invalid ILT algorithm "{}". Must be one of {}'.format(
                ilt_algorithm, '{"' + '", "'.join(ILT_ALGORITHMS.keys()) + '"}.'
            )
        )
    if not ilt_reconstruction_terms % 2:
        raise ValueError(
            'Invalid "ilt_reconstruction_terms", must be an odd input number. Was given an even number of {}'.format(
                ilt_reconstruction_terms
            )
        )
    if options is None:
        options = {}
    else:
        options = options.copy()
    to = torch.squeeze(t)
    if len(to.shape) >= 2:
        raise NotImplementedError(
            "Handling (MiniBatchSize, SeqLen) is not implemented yet. Coming soon. If urgent post an issue on GitHub"
        )
    if len(p.shape) >= 3:
        p = torch.squeeze(p)
    if recon_dim is None:
        recon_dim = p.shape[1]
    return options, recon_dim, p



import os
import re
import setuptools


def read(fname: str) -> str:
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def find_version() -> str:
    version_file = read("torchlaplace/__init__.py")
    return version_file.split('__version__')[1].split('"')[1]


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

required = [
    "torch>=1.3.0",
    "scipy>=1.4.0",
]

extras = {
    "experiments": [
        "ddeint",
        "keyboard",
        "matplotlib",
        "pandas",
        "sklearn",
        "stribor==0.1.0",
        "torchdiffeq>=0.2.1",
        "torchvision",
        "tqdm",
        "Sphinx",
        "sphinx-rtd-theme",
        "sphinx-panels",
        "nbsphinx"
    ]
}

extras['all'] = list(set([item for group in extras.values() for item in group]))



setuptools.setup(
    name="torchlaplace",
    version=find_version(),
    author="Sam Holt",
    author_email="samuel.holt.direct@gmail.com",
    description="Differentiable Laplace Reconstructions in PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/samholt/NeuralLaplace",
    license="MIT",
    packages=setuptools.find_packages(),
    python_requires="~=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=required,
    extras_require=extras,
)
